@@ -48,44 +48,45 @@ extern int	errno;
 
 static void gpm_killed(int);
 
-/*-------------------------------------------------------------------*/
-char *opt_type=DEF_TYPE;
+/*
+ * all the values duplicated for dual-mouse operation are
+ * now in this structure (see gpmInt.h)
+ */
+
+struct mouse_features mouse_table[3] = {
+  {
+  DEF_TYPE, DEF_DEV, DEF_SEQUENCE,
+  DEF_BAUD, DEF_SAMPLE, DEF_DELTA, DEF_ACCEL, DEF_SCALE,
+  DEF_TIME, DEF_CLUSTER, DEF_THREE, DEF_TOGGLE,
+  (Gpm_Type *)NULL
+  },
+};
+struct mouse_features *which_mouse;
+
+/* These are only the 'global' options */
+
 char *opt_lut=DEF_LUT;
-char *opt_dev=DEF_DEV;
-char *opt_sequence=DEF_SEQUENCE;
-int opt_baud=DEF_BAUD;
-int opt_sample=DEF_SAMPLE;
-int opt_delta=DEF_DELTA;
-int opt_accel=DEF_ACCEL;
-int opt_scale=DEF_SCALE;
-int opt_time=DEF_TIME;
-int opt_cluster=DEF_CLUSTER;
-int opt_three=DEF_THREE;
-int opt_dirty=DEF_DIRTY;
 int opt_test=DEF_TEST;
 int opt_ptrdrag=DEF_PTRDRAG;
 int opt_kill=0;
-int opt_toggle=DEF_TOGGLE;
-
+int opt_repeater=0, opt_double=0;
 static int opt_resize=0; /* not really an option */
 
 char *prgname;
-Gpm_Type *m_type;
 struct winsize win;
 int maxx, maxy;
+int fifofd;
 
 int eventFlag=0;
 Gpm_Cinfo *cinfo[MAX_VC+1];
 fd_set selSet, readySet, connSet;
 
-/* these are needed to provide a fictious motion to the client */
-int mouseX, mouseY;
-
 /*===================================================================*/
 /*
  *      first, all the stuff that used to be in gpn.c 
  */
 /*-------------------------------------------------------------------*/
+/*-------------------------------------------------------------------*/
 static inline int open_console(const int mode)
 {
 int fd;
@@ -96,28 +97,27 @@ int fd;
 }
 
 /*-------------------------------------------------------------------*/
-static inline int check_mode(int *fdptr)
+static inline int wait_text(int *fdptr)
 {
-int fd, ch = 0;
-long kd_mode;
+int fd;
+int kd_mode;
 
-  while(1)
+  close(*fdptr);
+  do
     {
+    sleep(2);
     fd = open_console(O_RDONLY);
     if (ioctl(fd, KDGETMODE, &kd_mode)<0)
       oops("ioctl(KDGETMODE)");
     close(fd);
-    if (kd_mode == KD_TEXT) break;
-    /* should I close the mouse descriptor, now? */
-    if (!ch) close(*fdptr);
-    sleep(2); ch++;
     }
+  while (kd_mode != KD_TEXT);
 
-  if (ch) {
-    if ((*fdptr=open(opt_dev,O_RDWR))<0)
-      oops(opt_dev);
-    if (m_type->init)
-      m_type=(m_type->init)(*fdptr, m_type->flags, m_type);
+  /* reopen, reinit */
+  if ((*fdptr=open(opt_dev,O_RDWR))<0)
+    oops(opt_dev);
+  if (m_type->init)
+    m_type=(m_type->init)(*fdptr, m_type->flags, m_type);
 
   if (opt_toggle)
     {
@@ -127,9 +127,8 @@ long kd_mode;
     modem_lines &= ~opt_toggle;
     ioctl(fd, TIOCMSET, &modem_lines);
     }
-  }
 
-  return (ch>0);
+  return (1);
 }
 
 /*-------------------------------------------------------------------*/
@@ -153,12 +152,13 @@ int fd;
 
   if ((fd=open_console(O_WRONLY))<0)
     oops("open_console");
-  DEBUGG((stderr,"ctl %i, mode %i\n",(int)*buf,arg[4]));
+  PDEBUGG((stderr,"ctl %i, mode %i\n",(int)*buf,arg[4]));
   if (ioctl(fd, TIOCLINUX, buf+sizeof(short)-1) < 0)
     oops("ioctl(TIOCLINUX)");
   close(fd);
 }
 
+
 /*-------------------------------------------------------------------*/
 static inline void selection_paste(void)
 {
@@ -212,7 +212,6 @@ static int x1=1, y1=1, x2, y2;
 	  return 0;
 
 	case GPM_B_MIDDLE:
-	  if (!opt_three) opt_three++;
 	  selection_paste();
 	  return 0;
 
@@ -256,11 +255,11 @@ MAGIC_P((static int magic=GPM_MAGIC));
 /*-------------------------------------------------------------------*/
 /*-------------------------------------------------------------------*/
 /*-------------------------------------------------------------------*/
-static inline char *getMouseData(int fd, Gpm_Type *m_type)
+static inline char *getMouseData(int fd, Gpm_Type *type)
 {
 static unsigned char data[32]; /* quite a big margin :) */
-char *edata=data+m_type->packetlen;
-int howmany=m_type->howmany;
+char *edata=data+type->packetlen;
+int howmany=type->howmany;
 int i,j;
 /*....................................... read and identify one byte */
 
@@ -275,6 +274,7 @@ int i,j;
       if (m_type->getextra)
 	{
 	data[1]=GPM_EXTRA_MAGIC_1; data[2]=GPM_EXTRA_MAGIC_2;
+	LOG(("Extra %02x",data[0]));
 	return data;
 	}
       LOG(("Error in protocol"));
@@ -307,22 +307,28 @@ int i,j;
       LOG(("Skipping a data packet (?)"));
       return NULL;
       }
+  LOG(("Data %02x %02x %02x (%02x)",data[0],data[1],data[2],data[3]));
   return data;
 }
 
+
+static int statusX,statusY,statusB; /* to return info */
+static int statusC=0; /* clicks */
+
 /*-------------------------------------------------------------------*/
-static inline int processMouse(int fd, Gpm_Event *event, Gpm_Type *m_type)
+static inline int processMouse(int fd, Gpm_Event *event, Gpm_Type *type,
+			       int kd_mode)
 {
 char *data;
 static int fine_dx, fine_dy;
 static int i, margin;
 static Gpm_Event nEvent;
 static struct vt_stat stat;
-static int clicks=0;    /* 0,1 or 2 */
 static struct timeval tv1={0,0}, tv2; /* tv1=0: force first click as single */
 static struct timeval timeout={0,0};
 fd_set fdSet;
 static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
+/* static int buttonlock, buttonlockflag; */
 
 #define GET_TIME(tv) (gettimeofday(&tv, (struct timezone *)NULL))
 #define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000+ \
@@ -357,7 +363,7 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
 
       if (oldB!=newB)
 	{
-	eventFlag = (i!=0);
+	eventFlag = (i!=0)*(which_mouse-mouse_table); /* 1 or 2 */
 	break;
 	}
 
@@ -374,6 +380,29 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
     
     } /* eventFlag */
 
+/*....................................... update the button number */
+
+  if ((event->buttons&GPM_B_MIDDLE) && !opt_three) opt_three++;
+
+/*....................................... we're a repeater, aren't we? */
+
+  if (kd_mode!=KD_TEXT)
+    {
+    unsigned char buffer[5];
+
+    /* sluggish... */
+    buffer[0]=(event->buttons ^ 0x07) | 0x80;
+    buffer[1]=(event->dx>>1) &0xFF;
+    buffer[2]=-(event->dy>>1) &0xFF;
+    buffer[3]=buffer[1] + (event->dx & 1);
+    buffer[4]=buffer[2] - (event->dy & 1);
+    write(fifofd,buffer,5);
+    
+    return 0; /* no events nor information for clients */
+    }
+
+/*....................................... no, we arent a repeater, go on */
+
   /* use fine values now */
   fine_dx+=event->dx;             fine_dy+=event->dy;
   event->dx=fine_dx/opt_scale;    event->dy=fine_dy/opt_scale;
@@ -385,6 +414,7 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
 /*....................................... fill missing fields */
 
   event->x+=event->dx, event->y+=event->dy;
+  statusB=event->buttons;
 
   i=open_console(O_RDONLY);
   ioctl(i,VT_GETSTATE,&stat);
@@ -404,14 +434,13 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
     case GPM_DOWN:
       GET_TIME(tv2);
       if (tv1.tv_sec && (DIF_TIME(tv1,tv2)<opt_time)) /* check first click */
-	clicks++, clicks%=3; /* 0, 1 or 2 */
-      else clicks=0;
+	statusC++, statusC%=3; /* 0, 1 or 2 */
+      else statusC=0;
       break;
 
     case GPM_UP:
       GET_TIME(tv1);
       event->buttons^=oldB; /* for button-up, tell which one */
-      event->type|= (GPM_SINGLE<<clicks);
       event->type|= (oldT&GPM_MFLAG);
       break;
 
@@ -420,11 +449,12 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
       break;
 
     case GPM_MOVE:
-      clicks=0;
+      statusC=0;
     default:
       break;
     }
-  event->clicks=clicks;
+  event->clicks=statusC;
+  event->type|=(GPM_SINGLE<<statusC);
 
 /* UGLY */
 /* The current policy is to force the following behaviour:
@@ -449,7 +479,7 @@ static int newB=0, oldB=0, oldT=0; /* old buttons and Type to chain events */
                                       event->buttons, event->vc,
                                       event->clicks));
   /* update the global state */
-  mouseX=event->x; mouseY=event->y;
+  statusX=event->x; statusY=event->y;
 
   return 1;
 }
@@ -487,8 +517,11 @@ static inline int processRequest(int fd, int vc)
 {
 int i;
 Gpm_Cinfo *cinfoPtr, *next;
+Gpm_Connect conn;
+static Gpm_Event event;
+static struct vt_stat stat;
 
-  LOG(("Request on %i (console %i)",fd,vc));
+  PDEBUGG((stderr,"Request on %i (console %i)\n",fd,vc));
 
   /* Hmm... data on inactive client: drop him */
   if (vc==-1)                   
@@ -501,6 +534,7 @@ Gpm_Cinfo *cinfoPtr, *next;
 	  LOG(("Closing"));
 	  close(fd);
 	  FD_CLR(fd,&connSet);
+	  FD_CLR(fd,&readySet);
 	  next=cinfoPtr->next;
 	  cinfoPtr->next=next->next;
 	  free(next);
@@ -509,20 +543,51 @@ Gpm_Cinfo *cinfoPtr, *next;
 
   if (vc==-1) oops("Unknown fd selected");
 
-  i=get_data(&(cinfo[vc]->data),fd);
+  i=get_data(&conn,fd);
 
-  if (!i)
+  if (!i) /* no data */
     {
     LOG(("Closing"));
     close(fd);
     FD_CLR(fd,&connSet);
+    FD_CLR(fd,&readySet);
     cinfoPtr=cinfo[vc];
     cinfo[vc]=cinfoPtr->next; /* pop the stack */
     free(cinfoPtr);
     return(-1);
     }
+  if (i==-1)
+    return -1; /* too few bytes */
 
-  return (i ? 0 : -1);
+  if (conn.pid!=0)
+    {
+    cinfo[vc]->data = conn;
+    return 0;
+    }
+
+  /* Aha, request for information (so-called snapshot) */
+
+  switch(conn.vc)
+    {
+    case GPM_REQ_SNAPSHOT:
+        i=open_console(O_RDONLY);
+	ioctl(i,VT_GETSTATE,&stat);
+	event.modifiers=6; /* code for the ioctl */
+	if (ioctl(i,TIOCLINUX,&(event.modifiers))<0)
+	  oops("get_shift_state");
+	close(i);
+	event.vc = stat.v_active;
+        event.x=statusX;   event.y=statusY;
+        event.dx=maxx;     event.dy=maxy;
+        event.buttons= statusB;
+        event.clicks=statusC;
+
+    case GPM_REQ_BUTTONS:
+        event.type= (opt_three==1 ? 3 : 2); /* buttons */
+        write(fd,&event,sizeof(Gpm_Event));
+      }
+
+  return 0;
 }
 
 /*-------------------------------------------------------------------*/
@@ -578,7 +643,7 @@ struct sockaddr_un addr; /* reuse this each time */
 /* if the client gets motions, give it the current position */
   if(request->eventMask & GPM_MOVE)
     {
-    Gpm_Event event={0,0,vc,0,0,mouseX,mouseY,GPM_MOVE,0,0};
+    Gpm_Event event={0,0,vc,0,0,statusX,statusY,GPM_MOVE,0,0};
     do_client(info, &event);
     }
 
@@ -598,8 +663,9 @@ int i;
     win.ws_col=80; win.ws_row=25;
     }
   maxx=win.ws_col; maxy=win.ws_row;
-  DEBUGG((stderr,"%i - %i\n",maxx,maxy));
-  ePtr->x=maxx/2; ePtr->y=maxy/2;
+  PDEBUGG((stderr,"%i - %i\n",maxx,maxy));
+  statusX=ePtr->x=maxx/2; statusY=ePtr->y=maxy/2;
+  
 }
 
 /*-------------------------------------------------------------------*/
@@ -607,18 +673,22 @@ int main(int argc, char **argv)
 {
 int mousefd, ctlfd, newfd;
 struct sockaddr_un ctladdr;
-int i, len;
+int i, len, kd_mode;
 struct   timeval timeout;
 int maxfd=-1;
 int pending;
 Gpm_Event event;
 
   prgname=argv[0];
+  setuid(0); /* just in case... */
+
 /*....................................... parse command line */
 
   mousefd=cmdline(argc, argv);
   maxfd=max(mousefd,maxfd);
 
+/*....................................... catch interesting signals */
+
   signal(SIGTERM, gpm_killed);
   signal(SIGINT,  gpm_killed);
   signal(SIGUSR1, gpm_killed); /* usr1 is used by a new gpm killing the old */
@@ -641,6 +711,9 @@ Gpm_Event event;
       oops(ctladdr.sun_path);
   maxfd=max(maxfd,ctlfd);
 
+/* is this a bug in the new kernels? */
+  chmod(GPM_NODE_CTL,0777);
+
 /*....................................... get screen dimensions */
 
   get_console_size(&event);
@@ -655,41 +728,44 @@ Gpm_Event event;
   FD_ZERO(&connSet);
   FD_SET(ctlfd,&connSet);
 
+  if (opt_double)
+    FD_SET(mouse_table[2].fd,&connSet);
+
+  readySet=connSet;
+  FD_SET(mousefd,&readySet);
 
   signal(SIGPIPE,SIG_IGN);  /* WARN */
+
+/*--------------------------------------- main loop begins here */
+
   while(1)
     {
-    check_mode(&mousefd);
-    maxfd=max(maxfd,mousefd);
-
-    /*
-     * readySet is mouse and ctl, connSet is connections only. I must
-     * re-init readySet because mousefd can be changed.
-     */
-    readySet=connSet;
-    FD_SET(mousefd,&readySet);
     selSet=readySet;
     resetTimeout();
     if (opt_test) timeout.tv_sec=0;
-    
-    if (eventFlag)
+
+    if (eventFlag) /* an event left over by clustering */
+      {
       pending=1;
+      FD_ZERO(&selSet);
+      FD_SET(mouse_table[eventFlag].fd,&selSet);
+      }
     else
       while ((pending=select(maxfd+1,&selSet, NULL_SET,NULL_SET,&timeout))==0)
 	{
-	if (check_mode(&mousefd))
-	  break;
 	LOGG(("Hi"));
 	selSet=readySet;
 	resetTimeout();
 	} /* go on */
 
-    if (opt_resize)
+
+    if (opt_resize) /* did the console resize? */
       {
       get_console_size(&event);
       opt_resize--;
       signal(SIGWINCH,gpm_killed); /* reinstall handler */
 
+      /* and notify clients */
       for(i=0; i<MAX_VC+1; i++)
 	if (cinfo[i])
 	  kill(cinfo[i]->data.pid,SIGWINCH);
@@ -697,6 +773,7 @@ Gpm_Event event;
 
     if (pending<0)
       {
+      if (errno==EBADF) oops("select()");
       LOG(("select(): %s",strerror(errno)));
       selSet=readySet;
       resetTimeout();
@@ -705,21 +782,49 @@ Gpm_Event event;
 
     LOGG(("selected %i times",pending));
 
-/*....................................... got mouse, process event */
+/*....................................... manage graphic mode */
 
-    if (FD_ISSET(mousefd,&selSet) || eventFlag)
+    /* 
+     * Be sure to be in text mode. This used to be before select,
+     * but actually it only matters if you have events.
+     */
+    {
+    int fd = open_console(O_RDONLY);
+    if (ioctl(fd, KDGETMODE, &kd_mode)<0)
+      oops("ioctl(KDGETMODE)");
+    close(fd);
+    if (kd_mode != KD_TEXT && !opt_repeater)
       {
-      FD_CLR(mousefd,&selSet); pending--;
-      LOGG(("Got mouse"));
-      if (processMouse(mousefd, &event, m_type))
-	/*
-	 * pass it to the client, if any
-	 * or to the default handler, if any
-	 * or to the selection handler
-	 */
-        (cinfo[event.vc] && do_client(cinfo[event.vc], &event))
-        || (cinfo[0]     && do_client(cinfo[0], &event))
-        || do_selection(&event);
+      wait_text(&mousefd);
+      maxfd=max(maxfd,mousefd);
+      readySet=connSet;
+      FD_SET(mousefd,&readySet);
+      continue; /* reselect */
+      }
+    }
+/*....................................... got mouse, process event */
+/*
+ * Well, actually, run a loop to mantain inlining of functions without
+ * lenghtening the file. This is not too clean a code, but it works....
+ */
+
+    for (i=1; i <= 1+opt_double; i++)
+      {
+      which_mouse=mouse_table+i; /* used to access options */
+      if (FD_ISSET(which_mouse->fd,&selSet))
+	  {
+	  FD_CLR(which_mouse->fd,&selSet); pending--;
+	  LOGG(("Got mouse"));
+	  if (processMouse(which_mouse->fd, &event, m_type, kd_mode))
+	    /*
+	     * pass it to the client, if any
+	     * or to the default handler, if any
+	     * or to the selection handler
+	     */
+	    (cinfo[event.vc] && do_client(cinfo[event.vc], &event))
+	       || (cinfo[0]        && do_client(cinfo[0],        &event))
+	       ||  do_selection(&event);
+	  }
       }
 
 /*....................................... got connection, process it */
@@ -727,11 +832,12 @@ Gpm_Event event;
     if (pending && FD_ISSET(ctlfd,&selSet))
       {
       FD_CLR(ctlfd,&selSet); pending--;
-      LOG(("Got connection"));
+      PDEBUGG((stderr,"Connection\n"));
       newfd=processConn(ctlfd);
       if (newfd>=0)
 	{
 	FD_SET(newfd,&connSet);
+	FD_SET(newfd,&readySet);
 	maxfd=max(maxfd,newfd);
         }
       }
@@ -742,6 +848,7 @@ Gpm_Event event;
       if (cinfo[i] && FD_ISSET(cinfo[i]->fd,&selSet))
 	{
 	int fd=cinfo[i]->fd;
+	PDEBUGG((stderr,"Request\n"));
         FD_CLR(fd,&selSet); pending--;
 	if ((processRequest(fd,i)==-1) && maxfd==fd)
 	  maxfd--; /* WARN */
@@ -768,15 +875,15 @@ static void gpm_killed(int signo)
 {
   if (signo==SIGWINCH)
     {
-    DEBUGG((stderr,"%s pid %i is resizing :-)\n",prgname,getpid()));
+    PDEBUGG((stderr,"%s pid %i is resizing :-)\n",prgname,getpid()));
     opt_resize++;
     return;
     }
   if (signo==SIGUSR1)
-    fprintf(stderr,"%s pid %i killed by a new %s\n",prgname,getpid(),prgname);
+    PDEBUG((stderr,"%s pid %i killed by a new %s\n",prgname,getpid(),prgname));
   else
     {
-    DEBUG((stderr,"Removing files...\n"));
+    PDEBUGG((stderr,"Removing files...\n"));
     unlink(GPM_NODE_PID);
     unlink(GPM_NODE_CTL);
     }
