@@ -19,13 +19,12 @@
  ********/
 
 #define GPM_NAME    "gpm-Linux"
-#define GPM_RELEASE "0.98"
-#define GPM_DATE    "March 1995"
+#define GPM_RELEASE "1.00"
+#define GPM_DATE    "July 1995"
 
-/* You can define DEBUG, DEBUGG, or neither. */
-/* Well, better putting them in the Makefile... */
+/* You can define DEBUG, or undefine it. */
+/* Well, better putting it in the Makefile... */
 /* #undef DEBUG  */
-/* #undef  DEBUGG */
 
 /* Is logging enabled? */
 #undef CONFIG_GPM_LOG
@@ -54,7 +53,6 @@
 #define DEF_CLUSTER         10    /* maximum number of clustered events */
 #define DEF_THREE            0    /* have three buttons? */
 
-#define DEF_DIRTY            0
 #define DEF_TEST             0
-#define DEF_PTRDRAG         10    /* never */
+#define DEF_PTRDRAG          1    /* double or triple click */
 #define DEF_TOGGLE           0
