@@ -37,6 +37,11 @@ extern int Log(char *fmt, ...);
 /* a placeholder */
 #define LOGG(args) 
 
+/*....................................... Strange requests (iff conn->pid==0)*/
+
+#define GPM_REQ_SNAPSHOT 0
+#define GPM_REQ_BUTTONS  1
+
 /*....................................... Structures */
 
 
@@ -45,14 +50,16 @@ extern int Log(char *fmt, ...);
  */
 typedef struct Gpm_Type {
   char *name;
+  char *desc;             /* a descriptive line */
+  char *syn;              /* the XFree name     */
   int (*fun)(Gpm_Event *state, unsigned char *data);
   struct Gpm_Type *(*init)(int fd, unsigned short flags,
 						   struct Gpm_Type *type);
   unsigned short flags;
   unsigned char proto[4];
   int packetlen;
-  int howmany; /* howmany bytes to read at a time */
-  int getextra; /* does it get an extra byte? (only mouseman) */
+  int howmany;            /* howmany bytes to read at a time */
+  int getextra;           /* does it get an extra byte? (only mouseman) */
 }                   Gpm_Type;
 
 #define GPM_EXTRA_MAGIC_1 0xAA
@@ -65,25 +72,45 @@ typedef struct Gpm_Cinfo {
 }              Gpm_Cinfo;
 
 
-/*....................................... Compatibility issues */
-
-#ifdef EWOULDBLOCK
-#  define GPM_AGAIN EAGAIN
-#else
-#  define GPM_AGAIN EAGAIN
-#endif
-
 /*....................................... Global variables */
 
+/* this structure is used to hide the dual-mouse stuff */
+
+struct mouse_features {
+  char *opt_type, *opt_dev, *opt_sequence;
+  int opt_baud,opt_sample,opt_delta, opt_accel, opt_scale;
+  int opt_time, opt_cluster, opt_three, opt_toggle;
+  Gpm_Type *m_type;
+  int fd;
+};
+
+extern struct mouse_features mouse_table[3], *which_mouse; /*the current one*/
+
+#define opt_type     (which_mouse->opt_type)
+#define opt_dev      (which_mouse->opt_dev)
+#define opt_sequence (which_mouse->opt_sequence)
+#define opt_baud     (which_mouse->opt_baud)
+#define opt_sample   (which_mouse->opt_sample)
+#define opt_delta    (which_mouse->opt_delta)
+#define opt_accel    (which_mouse->opt_accel)
+#define opt_scale    (which_mouse->opt_scale)
+#define opt_time     (which_mouse->opt_time)
+#define opt_cluster  (which_mouse->opt_cluster)
+#define opt_three    (which_mouse->opt_three)
+#define opt_toggle   (which_mouse->opt_toggle)
+
+#define m_type       (which_mouse->m_type)
+
+/* the other variables */
+
 extern char *prgname;
-extern char *opt_type, *opt_lut, *opt_dev;
-extern char *opt_sequence;
-extern int opt_baud, opt_sample, opt_delta, opt_accel, opt_scale;
-extern int opt_dirty, opt_test, opt_time, opt_cluster, opt_three, opt_ptrdrag;
-extern int opt_kill, opt_toggle;
+extern char *opt_lut;
+extern int opt_test, opt_ptrdrag;
+extern int opt_kill;
+extern int opt_repeater, opt_double;
+extern int fifofd;
 
 extern Gpm_Type mice[];
-extern Gpm_Type *m_type;
 extern struct winsize win;
 extern int maxx, maxy;
 extern Gpm_Cinfo *cinfo[MAX_VC+1];
@@ -97,6 +124,9 @@ int Log(char *fmt,...);              /* the logging utility               */
 int _oops(char *f, int n, char *s,int err);  /* printf's and exit(1)      */
 #define oops(s) _oops(__FILE__, __LINE__,(s),errno)
 int cmdline(int argc, char **argv);
+int giveInfo(int request, int fd);
+       /* mice.c */
+extern int M_listTypes(void);
 
 /*....................................... Dirty hacks */
 
@@ -108,3 +138,4 @@ int cmdline(int argc, char **argv);
 #else
 #define MAGIC_P(code) 
 #endif
+
