@@ -24,6 +24,10 @@
 #include "gpmInt.h"
 
 Gpm_Roi *gpm_roi=NULL;
+Gpm_Roi *gpm_current_roi=NULL;
+Gpm_Handler *gpm_roi_handler=NULL;
+void *gpm_roi_data;
+
 
 /*-------------------------------------------------------------------*/
 Gpm_Roi *Gpm_PushRoi(int x, int y, int X, int Y, int mask,
@@ -35,6 +39,10 @@ Gpm_Roi *new;
   new=(Gpm_Roi *)malloc(sizeof(Gpm_Roi));
   if (!new) return NULL;
 
+  /* use the roi handler, if still null */
+  if (!gpm_roi && !gpm_handler)
+	gpm_handler=Gpm_HandleRoi;
+
   new->xMin=x;         new->xMax=X;
   new->yMin=y;         new->yMax=Y;
   new->minMod=0;       new->maxMod=~0;
@@ -42,7 +50,7 @@ Gpm_Roi *new;
   new->eventMask=mask;
   new->owned=0;        /* use free() */
   new->handler=fun;
-  new->xtradata=xtradata;
+  new->clientdata = xtradata ? xtradata : (void *)new;
   return Gpm_RaiseRoi(new,NULL);
 }
 
@@ -52,6 +60,11 @@ Gpm_Roi *Gpm_UseRoi(Gpm_Roi *new)
   /* use a Roi by pushing it */
   new->prev      =     new->next=NULL;
   new->owned=1;        /* don't free() */
+
+  /* use the roi handler, if still null */
+  if (!gpm_roi && !gpm_handler)
+	gpm_handler=Gpm_HandleRoi;
+
   return Gpm_RaiseRoi(new,NULL);
 }
 
@@ -64,6 +77,8 @@ Gpm_Roi *Gpm_PopRoi(Gpm_Roi *which)
   if (gpm_roi==which) gpm_roi=which->next;
   
   if (which->owned==0) free(which);
+  if (gpm_current_roi==which)
+	gpm_current_roi=NULL;
   
   return gpm_roi; /* return the new top-of-stack */
 }
@@ -72,7 +87,10 @@ Gpm_Roi *Gpm_PopRoi(Gpm_Roi *which)
 Gpm_Roi *Gpm_RaiseRoi(Gpm_Roi *which, Gpm_Roi *before)
 {
   /* raise a Roi above another, or to top-of-stack */
+  if (!gpm_roi)
+	return gpm_roi=which;
   if (!before) before=gpm_roi;
+  if (before==which) return gpm_roi;
 
   if (which->prev) which->prev->next = which->next;
   if (which->next) which->next->prev = which->prev;
@@ -94,6 +112,7 @@ Gpm_Roi *Gpm_LowerRoi(Gpm_Roi *which, Gpm_Roi *after)
   if (!after)
     for (after=gpm_roi; after->next; after=after->next)
       ;
+  if (after==which) return gpm_roi;
   if (which->prev) which->prev->next = which->next;
   if (which->next) which->next->prev = which->prev;
   if (gpm_roi==which) gpm_roi=which->next;
@@ -113,9 +132,65 @@ Gpm_Roi *Gpm_LowerRoi(Gpm_Roi *which, Gpm_Roi *after)
  * It generates also GPM_ENTER and GPM_LEAVE events.
  */
 
-int Gpm_HandleRoi(Gpm_Event *ePtr)
-{return 0;}
+int Gpm_HandleRoi(Gpm_Event *ePtr, void *clientdata)
+{
+static Gpm_Event backEvent;
+Gpm_Roi *roi=gpm_current_roi;
 
+/*
+ * If motion or press, look for the interested roi.
+ * Drag and release will be reported to the old roi.
+ */
 
+  if (ePtr->type&(GPM_MOVE|GPM_DOWN))
+	for (roi=gpm_roi; roi; roi=roi->next)
+	  {
+	  if (roi->xMin > ePtr->x || roi->xMax < ePtr->x)
+		continue;
+	  if (roi->yMin > ePtr->y || roi->yMax < ePtr->y)
+		continue;
+	  if (roi->minMod & ePtr->modifiers < roi->minMod) continue;
+	  if (roi->maxMod & ePtr->modifiers < ePtr->modifiers) continue;
+	  break;
+	  }
 
+/*
+ * Now generate the leave/enter events
+ */
+
+  if (roi!=gpm_current_roi)
+	{
+	if (gpm_current_roi && (gpm_current_roi->eventMask & GPM_LEAVE))
+	  {
+	  backEvent.type=GPM_LEAVE;
+	  (*(gpm_current_roi->handler))(&backEvent,gpm_current_roi->clientdata);
+	  }
+	if (roi && (roi->eventMask & GPM_ENTER))
+	  {
+	  backEvent.type=GPM_ENTER;
+	  (*(roi->handler))(&backEvent,roi->clientdata);
+	  }
+	}
+  gpm_current_roi=roi;
+  /*
+   * events not requested are discarded
+   */
+  if (GPM_BARE_EVENTS(ePtr->type) & roi->eventMask == 0) return 0; 
+
+  backEvent=*ePtr; /* copy it, so the main one is unchanged */
+  if (!roi)
+	{
+	if (gpm_roi_handler)
+	  return (*gpm_roi_handler)(&backEvent,gpm_roi_data);
+	return 0;
+	}
+
+/*
+ * Ok, now report the event as it is, after modifying x and y
+ */
+
+  backEvent.x -= roi->xMin;
+  backEvent.y -= roi->yMin;
+  return (*(roi->handler))(&backEvent,roi->clientdata);
+}
 
