@@ -1,8 +1,8 @@
 /*
  * mice.c - mouse definitions for gpm-Linux
  *
- * Copyright 1993   ajh@gec-mrc.co.uk (Andrew Haylett)
- * Copyright 1994   rubini@ipvvis.unipv.it (Alessandro Rubini)
+ * Copyright 1993        ajh@gec-mrc.co.uk (Andrew Haylett)
+ * Copyright 1994,1995   rubini@ipvvis.unipv.it (Alessandro Rubini)
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -51,7 +51,7 @@
 
 /*========================================================================*/
 /*
- * Ok, here we are: first, provide the functions
+ * Ok, here we are: first, provide the functions.
  * The return value is the number of unprocessed bytes
  */
 
@@ -160,12 +160,13 @@ static int M_ps2(Gpm_Event *state,  unsigned char *data)
   return 0;
 }
 
+#define GPM_B_BOTH (GPM_B_LEFT|GPM_B_RIGHT)
 static int M_mman(Gpm_Event *state,  unsigned char *data)
 {
   /*
    * the damned MouseMan has 3/4 bytes packets. The extra byte 
    * is only there if the middle button is active.
-   * I get the extra byte as a packet with magic numebers in it.
+   * I get the extra byte as a packet with magic numbers in it.
    * and then switch to 4-byte mode.
    */
   static unsigned char prev=0;
@@ -175,7 +176,7 @@ static int M_mman(Gpm_Event *state,  unsigned char *data)
 	{
 	/* got unexpected fourth byte */
 	LOG(("Extra byte = %02x",*data));
-	if (*data != 0x20) return -1; /* must be 0x20 */
+	if ((*data & 0xfc) != 0x20) return -1; /* must be 0x20-0x23 */
 	state->buttons= (prev |= GPM_B_MIDDLE);
 	state->dx=state->dy=0;
 
@@ -196,6 +197,10 @@ static int M_mman(Gpm_Event *state,  unsigned char *data)
   /* motion is independent of packetlen... */
   state->dx= (char)(((data[0] & 0x03) << 6) | (data[1] & 0x3F));
   state->dy= (char)(((data[0] & 0x0C) << 4) | (data[2] & 0x3F));
+
+  /* This "chord-middle" behaviour was reported by David A. van Leeuwen */
+  if ( ((prev^state->buttons) & GPM_B_BOTH)==GPM_B_BOTH )
+	state->buttons = state->buttons ? GPM_B_MIDDLE : 0;
   prev=state->buttons;
   
   return 0;
@@ -277,7 +282,7 @@ fd_set set; struct timeval timeout={0,0}; /* used when not debugging */
   write(fd,sampletab[i].code,1);
 #endif
 
-#ifndef DEBUGGING
+#ifndef DEBUG
   /*
    * flush any pending input (thanks, Miguel)
    * moreover, damned mouseman's can be detected here (though it isn't)
@@ -293,13 +298,25 @@ fd_set set; struct timeval timeout={0,0}; /* used when not debugging */
 	  }
 	break;
 	}
-#endif
 
   if (type->fun==M_ms && i==2 && c==0x33) /* Aha.. a mouseman... */
 	{
 	LOG(("MouseMan detected"));
 	return mice; /* it is the first */
 	}
+#endif
+
+  /*
+   * reset the MouseMan/TrackMan to use the 3/4 byte protocol
+   * (Stephen Lee, sl14@crux1.cit.cornell.edu)
+   */
+  if (type->fun==M_mman)
+	{
+	setspeed(fd, 1200, 1200, flags);
+	write(fd, "*X", 2);
+	setspeed(fd, 1200, opt_baud, flags);
+	}
+
   return type;
 }
 
@@ -338,29 +355,62 @@ int busmouse;
 /* Finally, the table */
 #define STD_FLG (CREAD|CLOCAL|HUPCL)
 
+/*
+ * Note that mman must be the first, and ms the second (I use this info
+ * in mouse-test.c, as a quick and dirty hack
+ */
 Gpm_Type mice[]={
-  {"mman", M_mman, I_serial, CS7 | STD_FLG, /* MUST BE FIRST */
+  {"mman", "The \"MouseMan\" and similar devices (3/4 bytes per packet).",
+           "Mouseman", M_mman, I_serial, CS7 | STD_FLG, /* MUST BE FIRST */
                                 {0x40, 0x40, 0x40, 0x00}, 3, 1, 1},
-  {"ms",   M_ms, I_serial, CS7 | STD_FLG,
+  {"ms",   "The original micro$oft protocol, with a middle-button extension.",
+           "", M_ms, I_serial, CS7 | STD_FLG,
                                 {0x40, 0x40, 0x40, 0x00}, 3, 1, 0},
-  {"bare", M_bare, I_serial, CS7 | STD_FLG,
+  {"bare", "The ms protocol, unextended. Necessary for some 2-buttons mice.",
+           "Microsoft", M_bare, I_serial, CS7 | STD_FLG,
                                 {0x40, 0x40, 0x40, 0x00}, 3, 1, 0},
-  {"sun",  M_sun, I_serial, CS8 | CSTOPB | STD_FLG,
-                                {0xf8, 0x80, 0x00, 0x00}, 3, 1, 0},
-  {"msc",  M_msc, I_serial, CS8 | CSTOPB | STD_FLG,
+  {"msc",  "Mouse-Systems-Compatible. Used in most 3-button mice (5-bytes).",
+           "MouseSystems", M_msc, I_serial, CS8 | CSTOPB | STD_FLG,
                                 {0xf8, 0x80, 0x00, 0x00}, 5, 1, 0},
-  {"mm",   M_mm, I_serial, CS8 | PARENB|PARODD | STD_FLG,
+  {"sun",  "'msc' protocol, but only 3 bytes per packet.",
+           "", M_sun, I_serial, CS8 | CSTOPB | STD_FLG,
+                                {0xf8, 0x80, 0x00, 0x00}, 3, 1, 0},
+  {"mm",   "MM series. Probably an old protocol...",
+           "MMSeries", M_mm, I_serial, CS8 | PARENB|PARODD | STD_FLG,
                                 {0xe0, 0x80, 0x80, 0x00}, 3, 1, 0},
-  {"logi", M_logi, I_logi, CS8 | CSTOPB | STD_FLG,
+  {"logi", "Used in some Logitech devices (only serial).",
+           "Logitech", M_logi, I_logi, CS8 | CSTOPB | STD_FLG,
                                 {0xe0, 0x80, 0x80, 0x00}, 3, 3, 0},
-  {"bm",   M_bm, NULL,  STD_FLG,
+  {"bm",   "Micro$oft busmice and compatible devices.",
+           "BusMouse", M_bm, NULL,  STD_FLG,
                                 {0xf8, 0x80, 0x00, 0x00}, 3, 3, 0},
-  {"ps2",  M_ps2, NULL,  STD_FLG,
+  {"ps2",  "Busmice of the ps/2 series. Most busmice, actually.",
+           "PS/2", M_ps2, NULL,  STD_FLG,
                                 {0xc0, 0x00, 0x00, 0x00}, 3, 1, 0},
 
-  {"",     NULL, NULL, 0,       {0x00, 0x00, 0x00, 0x00}, 0, 0, 0}
+  {"", "", "", NULL, NULL, 0,   {0x00, 0x00, 0x00, 0x00}, 0, 0, 0}
 };
 
+/*------------------------------------------------------------------------*/
+/* and the help */
+
+int M_listTypes(void)
+{
+Gpm_Type *type;
+
+ printf("\n" GPM_NAME " " GPM_RELEASE ", " GPM_DATE "\n");
+ printf("Available mouse types are:\n\n");
+ printf("  name   synonym         description\n\n");
+ for (type=mice; type->fun; type++)
+   printf("  %-6s %-12s %s\n",type->name, type->syn, type->desc);
+
+ putchar('\n');
+
+return 1; /* to exit() */
+}
+
+
+
 
 
 
