@@ -1,7 +1,7 @@
 /*
  * rmev.c - reduced mev.
  *
- * Copyright 1994   rubini@ipvvis.unipv.it (Alessandro Rubini)
+ * Copyright 1994,1995   rubini@ipvvis.unipv.it (Alessandro Rubini)
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -48,8 +48,8 @@
 #endif
 
 #define RMEV_NAME "Mev"
-#define RMEV_RELEASE "0.98"
-#define RMEV_DATE "March 1995"
+#define RMEV_RELEASE "1.00"
+#define RMEV_DATE "July 1995"
 
 
 /*===================================================================*/
